/*
 * Provide access to keypad events on the configured device.
 *
 * (c) 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/dataspace.h>
#include "input-event-client.h"
#include "input-event-server.h"
#include "input-event-ops.h"
#include "input-keypad-client.h"

#include <l4/cxx/ipc_stream>
#include <l4/re/dataspace>
#include <l4/re/env>
#include <l4/re/event>
#include <l4/re/util/cap_alloc>
#include <l4/re/util/object_registry>
#include <l4/sys/capability>
#include <l4/sys/icu>
#include <l4/sys/kip.h>
#include <l4/sys/types.h>
#include <l4/util/util.h>

/* Handle invocations. */

int
Input_event_server::dispatch(l4_umword_t obj, L4::Ipc::Iostream &ios)
{
  l4_msgtag_t tag;
  l4_umword_t op;

  (void) obj;
  ios >> tag;

  switch (tag.label())
  {
    case L4::Meta::Protocol:
      return L4::Util::handle_meta_request<Input_event_interface>(ios);

    case L4RE_PROTO_EVENT:
      ios >> op;
      switch (op)
      {
        /* Reset the buffer state and return the buffer memory capability. */

        case Input_event_op_get_buffer:
          _events.reset();
          ios << _mem;
          return L4_EOK;

        default:
          return -L4_ENOSYS;
      }

    case L4_PROTO_IRQ:
      ios >> op;
      switch (op)
      {
        /* Just return the interrupt capability. */

        case Input_event_op_bind:
          ios << _irq;
          return L4_EOK;

        default:
          return -L4_ENOSYS;
      }

    default:
      return -L4_EBADPROTO;
  }
}



/* Event sending methods. */

void
Input_event_server::send_event(int type, int code, int value)
{
  L4Re::Event_buffer::Event event;

  event.time = l4_kip_clock(l4re_kip());
  event.payload.stream_id = 0;
  event.payload.type = type;
  event.payload.code = code;
  event.payload.value = value;

  /* Queue the event and trigger the interrupt. */

  _events.put(event);
  _irq->trigger();
}



static void handler(Input_event event, void *priv)
{
  Input_event_server *server = (Input_event_server *) priv;

  server->send_event(event.type, event.code, event.value);
}



static L4Re::Util::Registry_server<> server;



int run(void)
{
  Keypad *keypad = Keypad::get_keypad();
  Input_keypad_client client(keypad);
  void *buffer;

  // Memory allocation capability for the event data.

  L4::Cap<L4Re::Dataspace> mem = allocate_data(L4_PAGESIZE, &buffer);
  if (!mem.is_valid()) return 1;

  /* Obtain a capability for the interrupt. */

  L4::Cap<L4::Irq> irq = L4Re::Util::cap_alloc.alloc<L4::Irq>();
  if (!irq.is_valid()) return 1;

  /* Create an interrupt object. */

  if (l4_error(L4Re::Env::env()->factory()->create(irq))) return 1;

  // Event buffer for the data.

  L4Re::Event_buffer events(buffer, L4_PAGESIZE);

  // Initialise and register a server object.

  Input_event_server server_obj(mem, irq, events);
  server.registry()->register_obj(&server_obj, "ev");

  // Attach the event handler and wait for events.

  client.attach(handler, (void *) &server_obj);

  // Enter the IPC server loop.

  server.loop();
  return 0;
}
